#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import collections
import typing

T = typing.TypeVar('T', bound=typing.Any)


def to_list(x: T) -> list[T]:
    if isinstance(x, list):
        return x
    # noinspection PyUnresolvedReferences,PyProtectedMember
    if isinstance(x, collections.abc.Iterable) and not isinstance(x, str | bytes | typing._GenericAlias):
        return list(x)
    return [x]



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import Any


def unflatten_dict(flatten_dict: dict[str, Any], separator: str = '.') -> dict[str, Any | dict[str, Any]]:
    result: dict[str, Any | dict[str, Any]] = {}
    for key, value in flatten_dict.items():
        container = result
        path = key.split(separator)
        for step in path[:-1]:
            container = container.setdefault(step, {})
        container[path[-1]] = value
    return result



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from frozendict import frozendict


class FieldErrors(ValueError):
    def __init__(self, field_to_error: dict[str, Exception]) -> None:
        self.field_to_error = frozendict(field_to_error)


class MissingField(ValueError):
    pass


class UnsupportedType(TypeError):
    pass



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from ._exceptions import FieldErrors, MissingField, UnsupportedType
from .get_field_value import get_field_value, convert
from .convert_to_type import convert_to_type
from .to_dataclass import to_dataclass, dict_to_dataclass
from .get_converter import get_converter
from .get_dataclass_field_name_to_field import get_dataclass_field_name_to_field



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import collections.abc
import types
from typing import Any, Sequence, Union


def is_mapping(value: Any, mapping_args: Sequence[type | types.GenericAlias]) -> bool:
    if not isinstance(value, collections.abc.Mapping):
        return False

    if not mapping_args:
        return True

    if len(mapping_args) != 2:
        raise TypeError(f"Mapping requires exactly 2 type args, got: {mapping_args}")

    key_type_list = [mapping_args[0]]
    value_type_list = [mapping_args[1]]

    from .is_instance import is_instance
    for key, val in value.items():
        if not is_instance(key, key_type_list) or not is_instance(val, value_type_list):
            return False
    return True



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
import typing

from frozendict import frozendict

from .._type import DataclassProtocol, WideType, get_evaled_dataclass_fields


@functools.cache
def get_dataclass_field_name_to_type(dataclass_: type[DataclassProtocol]) -> frozendict[str, WideType]:
    result: dict[str, WideType] = typing.get_type_hints(dataclass_)
    # noinspection PyTypeChecker
    return frozendict(result | get_evaled_dataclass_fields(dataclass_))



#  Copyright (C) 2025
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import dataclasses
import types
import typing

from ._alias import MaybeAnnotated, WideType

# Type = typing.TypeVar("Type", bound=type)
# Type = typing.TypeVar("Type", bound=type | types.GenericAlias | types.UnionType)
Type = typing.TypeVar("Type", bound=WideType)


def try_extract_type_notes(value: MaybeAnnotated[Type]) -> tuple[Type, tuple]:
    if typing.get_origin(value) != typing.Annotated:
        return value, tuple()
    args = typing.get_args(value)
    return args[0], args[1:]



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import contextlib
import dataclasses
import inspect
import sys
import typing
from typing import Mapping, Any

from frozendict import frozendict

from .dataclass_protocol import DataclassProtocol
from ._alias import WideType

# noinspection PyUnresolvedReferences,PyProtectedMember
eval_type = typing._eval_type


@contextlib.contextmanager
def eval_forward_refs_in_local_dataclasses():
    yield
    current_frame = inspect.currentframe()
    contextmanager_frame = current_frame.f_back
    callee_frame = contextmanager_frame.f_back
    targets = [val for val in callee_frame.f_locals.values() if isinstance(val, DataclassProtocol)]
    _resolve_reverse_refs(*targets, localns=callee_frame.f_locals, globalns=callee_frame.f_globals)


def _resolve_reverse_refs(
        *targets: type[DataclassProtocol],
        localns: Mapping[str, Any], globalns: Mapping[str, Any]
) -> None:
    for target in targets:
        annotations = target.__annotations__
        for key, old_type in annotations.items():
            new_type = eval_type(old_type, globalns=globalns, localns=localns)
            if old_type != new_type:
                annotations[key] = new_type

        for f in dataclasses.fields(target):
            new_type = eval_type(f.type, globalns=globalns, localns=localns)
            if f.type != new_type:
                f.type = new_type


def get_evaled_dataclass_fields(obj: type[DataclassProtocol]) -> frozendict[str, WideType]:
    key_to_type = obj.__dict__.get('__key_to_type__')
    if key_to_type is None:
        # noinspection PyTypeChecker
        eval_dataclass_fields(obj)
        key_to_type = {f.name: f.type for f in dataclasses.fields(obj)}
        obj.__key_to_type__ = frozendict(key_to_type)
    return obj.__key_to_type__


def eval_dataclass_fields(obj: type[DataclassProtocol]) -> None:
    for base in reversed(obj.__mro__):
        base_globals = getattr(sys.modules.get(base.__module__, None), '__dict__', {})
        base_locals = dict(vars(base))
        if not isinstance(base, DataclassProtocol):
            continue
        fields = dataclasses.fields(base)
        for field in fields:
            value = field.type
            if value is None:
                value = type(None)
            if isinstance(value, str):
                value = typing.ForwardRef(value, is_argument=False, is_class=True)
            try:
                value = eval_type(value, base_globals, base_locals)
            except NameError as e:
                raise TypeError(
                    f'Failed to evaluate type {value!r}: not found referenced {e.name!r}. '
                    f'If it is class, created in local scope, wrap it in '
                    f'`with eval_forward_refs_in_local_dataclasses():` to evaluate type references in local scope '
                ) from None
            field.type = value



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import itertools
import types
from typing import Any, Iterable, Union


def is_tuple(value: Any, tuple_args: Iterable[type | types.GenericAlias]) -> bool:
    if not isinstance(value, tuple):
        return False

    has_ellipsis = False
    type_checks = []
    for inner_type in tuple_args:
        if inner_type is ...:
            has_ellipsis = True
        else:
            if has_ellipsis:
                raise TypeError("Ellipsis (three dots (...)) should be the last"
                                " in tuple to make it variable-length")
            type_checks.append(inner_type)
    if has_ellipsis:
        type_checks = itertools.cycle(type_checks)
    if not has_ellipsis and type_checks:
        if len(type_checks) != len(value):
            return False

    from .is_instance import is_instance
    for val, inner_type in zip(value, type_checks):
        if not is_instance(val, [inner_type]):
            return False

    return True



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import collections.abc


def get_container_type(type_origin) -> type:
    if not getattr(type_origin, "__abstractmethods__", None):
        return type_origin
    container_type = ORIGIN_TO_CONTAINER_TYPE.get(type_origin)
    if container_type is None:
        raise TypeError(f'Attempt to get_container_type from unexpected origin: {type_origin}')

    return container_type


ORIGIN_TO_CONTAINER_TYPE: dict[collections.abc.Container, type] = {
    collections.abc.Mapping: dict,
    collections.abc.Collection: list,
    collections.abc.Set: set,
}



#  Copyright (C) 2026
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from typing import Any

from .str_to_bool import str_to_bool


def to_bool(value: Any) -> bool:
    if isinstance(value, bool):
        return value
    if isinstance(value, int):
        return bool(value)
    if isinstance(value, str):
        return str_to_bool(value)
    raise TypeError(f'Attempt to convert {type(value).__qualname__} to bool: {value=}')
